#include <ros/ros.h>

// opencv image processing libraries
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/opencv.hpp>
#include "opencv2/objdetect/objdetect.hpp"
#include <geometry_msgs/Twist.h>

// to facilitate the processing of the images in ROS
#include <image_transport/image_transport.h> // for publishing and subscribing to images in ROS
#include <cv_bridge/cv_bridge.h> // to convert between ROS and OpenCV Image formats
#include <sensor_msgs/image_encodings.h>

#include <hog_haar_person_detection/Faces.h>
#include <hog_haar_person_detection/Pedestrians.h>
#include <hog_haar_person_detection/BoundingBox.h>

static const std::string OPENCV_WINDOW = "Image window";

class HogHaarPersonDetection
{
public:
  ros::NodeHandle nh_;
  image_transport::ImageTransport it_;
  image_transport::Subscriber im_sub_;
  image_transport::Publisher im_pub_;

  ros::Publisher faces_pub_;
  image_transport::Publisher line_pub;
  ros::Publisher pedestrians_pub_;

  cv::CascadeClassifier face_cascade_;

  cv::HOGDescriptor hog_;
  /* Minimum delay between 2 successive Twist commands. */
  static const uint16_t MIN_DELAY = 300;

  /* Maximum angle allowed before adjusting speed for taking corners. */
  static const uint8_t FW_CORNER_ALLOWED = 30;
  /* Maximum corner allowed to be taken. */
  static const uint8_t MAX_CORNER = 100;
  static const uint8_t KERNEL_SIZE = 3;
  static const uint8_t LOW_THR = 40;

  /* Direction constants. */
  static const uint8_t FORWARD = 0;
  static const uint8_t RIGHT = 1;
  static const int8_t LEFT = -1;

  /* Speed constants. */
  static const uint8_t FW_SPEED = 70;
  static const uint8_t TURN_SPEED = 45;

  /* Dimension constants. */
  static const int WIDTH = 1920;
  static const int HEIGHT = 1080;
  HogHaarPersonDetection() :
      it_(nh_)
  {
    // Get ROS params from launch file
    std::string image_topic;
    if (!nh_.getParam("image_topic", image_topic))
      ROS_ERROR("Could not get image_topic");

    std::string face_cascade_name_std;
    if (!nh_.getParam("face_cascade_name", face_cascade_name_std))
      ROS_ERROR("Could not get face_cascade_name");
    cv::String face_cascade_name = face_cascade_name_std;

    // Load the hog descriptor
    hog_.setSVMDetector(cv::HOGDescriptor::getDefaultPeopleDetector());

    // Load face detector HAAR cascades
    if (!face_cascade_.load(face_cascade_name))
      ROS_ERROR("--(!)Error loading face detector cascade \n");

    // Subscrive to input video feed and publish output video feed
    im_sub_ = it_.subscribe(image_topic, 1, &HogHaarPersonDetection::imageCallback, this);
    im_pub_ = it_.advertise("/camera_person_tracker/output_video", 1);

    // Publish detected faces and pedestrians.
    faces_pub_ = nh_.advertise<hog_haar_person_detection::Faces>("/person_detection/faces", 1000);
    pedestrians_pub_ = nh_.advertise<hog_haar_person_detection::Pedestrians>("/person_detection/pedestrians", 1000);
    line_pub = it_.advertise("/canny_edge/output_video", 1);
    cv::namedWindow(OPENCV_WINDOW);
  }

  ~HogHaarPersonDetection()
  {
    cv::destroyWindow(OPENCV_WINDOW);
  }
  cv::Vec4i findClosest(std::vector<cv::Vec4i>& lines)
  {
    cv::Vec4i closest;
    static const cv::Point ORIGIN(480, 960);
    /* Initialize the minimum distance to the largest value possible. */
    float min_dist = ORIGIN.y;
    float dist;

    /* Find the line that is the closest to the origin. */
    for (int i = 0; i < lines.size(); i++)
    {
      cv::Vec4i line = lines.at(i);
      dist = std::min(cv::norm(cv::Point2i(lines[i][0], lines[i][1]) - ORIGIN), cv::norm(cv::Point2i(lines[i][2], lines[i][3]) - ORIGIN));
      if (dist < min_dist)
      {
        min_dist = dist;
        closest = line;
      }
    }
    return closest;
  }
  void bestDirection(std::vector<cv::Vec4i>& lines, cv::Mat& img) {
      /* Initialize statically, less than ros::Time::now() because otherwise
       * the first 250ms are wasted. */
      static ros::Time last = ros::Time::now() - ros::Duration(1);

      /* Don't evaluate the line if the previous command was sent less than 250ms ago. */
      if ((ros::Time::now() - last).toSec() * 1000 < MIN_DELAY) {
          return;
      }

      cv::Vec4i c = findClosest(lines);
      /* Extract points from the closest line. */
      cv::Point2i p1(c[0], c[1]);
      cv::Point2i p2(c[2], c[3]);

      /* Draw the closest line. */
      cv::line(img, p1, p2, cv::Scalar(255,0,0), 3, CV_AA);

      /* Calculate the angle in degrees w.r.t the y-axis. */
      int angle = (atan2(c[3] - c[1], c[2] - c[0]) - CV_PI / 2) * 180 / CV_PI;

      /* Generate a Twist message. */
      if (angle > FW_CORNER_ALLOWED) {
          generateTwist(RIGHT, angle);
      } else if (angle < -FW_CORNER_ALLOWED) {
          generateTwist(LEFT, angle);
      } else {
          generateTwist(FORWARD, 0);
      }
  }
  void generateTwist(int dir, int angle) {
      /* Correct for angles going out of bounds. */
      if (angle < -MAX_CORNER)
          angle += 180;
      else if (angle > MAX_CORNER)
          angle -= 180;

      /* Create and publish a Twist message indicating the speed, angle and direction. */
      geometry_msgs::Twist twist;
      twist.angular.z = angle;
      /* Adjust speed if a corner is to be taken. */
      twist.linear.x = dir == FORWARD ? FW_SPEED : TURN_SPEED;
     // twist_pub.publish(twist);
  }
  void imageCallback(const sensor_msgs::ImageConstPtr& msg)
  {
    float xv;
    float yv;
    float angle;
    // Make image processable in ROS
    cv_bridge::CvImagePtr cv_ptr;
    cv_bridge::CvImage cvCanny;
    try
    {
      cv_ptr = cv_bridge::toCvCopy(msg, sensor_msgs::image_encodings::BGR8);
    }
    catch (cv_bridge::Exception& e)
    {
      ROS_ERROR("cv_bridge exception: %s", e.what());
      return;
    }
    cv::Mat im_bgr = cv_ptr->image;

    // HAAR Face detector
    std::vector<cv::Rect> detected_faces;
    cv::Mat im_gray;
    cv::cvtColor(im_bgr, im_gray, CV_BGR2GRAY);

    //line follower
    cv::Mat dst, detected_edges;
    //Reduce noise with a kernel 3x3
    cv::GaussianBlur(im_gray, dst, cv::Size(11, 11), 2);
    cv::Canny(dst, detected_edges, 80, 200, 3); //threshold1(80)和threshold2(123) 当中的小阈值用来控制边缘连接，大的阈值用来控制强边缘的初始分割。Sobel 算子内核大小(3).

    cvCanny.header = msg->header;
    cvCanny.encoding = sensor_msgs::image_encodings::TYPE_8UC1;
    cvCanny.image = detected_edges;

    line_pub.publish(cvCanny.toImageMsg());
    //std::vector<cv::Vec4i> lines;
    cv::Mat cdst;
    CvSeq* lines = 0;
    CvMemStorage* storage = cvCreateMemStorage(0);
    CvMat c_image = detected_edges;
    lines = cvHoughLines2(&c_image, storage, CV_HOUGH_PROBABILISTIC, 1, CV_PI / 180, 80, 50, 10);
    for (int i = 0; i < lines->total; i++)
    {
      CvPoint* line = (CvPoint*)cvGetSeqElem(lines, i);
      {
        xv = line[0].x - line[1].x;
        yv = line[0].y - line[1].y;
        angle = atan2(xv, yv) * 180 / 3.14159265;
        float buf = (line[0].x + line[1].x) / 2.0;

      }
    }
    //cv::HoughLinesP(detected_edges, lines, 1, CV_PI/180, 50, 50, 10 );
//     for( size_t i = 0; i < lines.size(); i++ )
//     {
//       cv::Vec4i l = lines[i];
//       cv::line(cdst, cv::Point(l[0], l[1]), cv::Point(l[2], l[3]), cv::Scalar(0,0,255), 3, CV_AA);
//     }

//     sensor_msgs::ImagePtr linemsg;
//     linemsg = cv_bridge::CvImage(std_msgs::Header(), "bgr8", cdst).toImageMsg();

// line_pub.publish(linemsg);

    /*
     cv::equalizeHist( im_gray, im_gray );
     face_cascade_.detectMultiScale(im_gray, detected_faces, 1.1, 2, 0|CV_HAAR_SCALE_IMAGE, cv::Size(30, 30));

     // Publish message of location and confidence of detected faces.
     // Also, draw detected faces to screen as circles.
     hog_haar_person_detection::Faces faces_msg;
     for(unsigned i = 0; i < detected_faces.size(); i++) {
     // Draw on screen.
     cv::Point center(detected_faces[i].x + detected_faces[i].width*0.5, detected_faces[i].y + detected_faces[i].height*0.5);
     cv::ellipse(im_bgr, center, cv::Size(detected_faces[i].width*0.5, detected_faces[i].height*0.5), 0, 0, 360, cv::Scalar(255, 0, 255), 4, 8, 0);

     // Add to published message.
     hog_haar_person_detection::BoundingBox face;
     face.center.x = detected_faces[i].x - detected_faces[i].width / 2;
     face.center.y = detected_faces[i].y - detected_faces[i].height / 2;
     face.width = detected_faces[i].width;
     face.height = detected_faces[i].height;
     faces_msg.faces.push_back(face);
     }
     faces_pub_.publish(faces_msg);

     // HOG pedestrian detector
     std::vector<cv::Rect> detected_pedestrian;
     hog_.detectMultiScale(im_gray, detected_pedestrian, 0.0, cv::Size(8, 8), cv::Size(0,0), 1.05, 4);

     // Publish message of location and confident of detected pedestrians.
     // Draw detections from HOG to the screen.
     hog_haar_person_detection::Pedestrians pedestrians_msg;
     for(unsigned i = 0; i < detected_pedestrian.size(); i++) {
     // Draw on screen.
     cv::rectangle(im_bgr, detected_pedestrian[i], cv::Scalar(255));

     // Add to published message.
     hog_haar_person_detection::BoundingBox pedestrian;
     pedestrian.center.x = detected_pedestrian[i].x - detected_pedestrian[i].width / 2;
     pedestrian.center.y = detected_pedestrian[i].y - detected_pedestrian[i].height / 2;
     pedestrian.width = detected_pedestrian[i].width;
     pedestrian.height = detected_pedestrian[i].height;
     pedestrians_msg.pedestrians.push_back(pedestrian);
     }
     pedestrians_pub_.publish(pedestrians_msg);

     // Update GUI Window
     cv::imshow(OPENCV_WINDOW, im_bgr);
     cv::waitKey(3);*/
  }
}
;

int main(int argc, char** argv)
{
  ros::init(argc, argv, "hog_haar_person_detection");
  HogHaarPersonDetection hhpd;
  ros::spin();
  return 0;
}
